{-# LANGUAGE OverloadedStrings #-}
module GEODE.Metadata.TSV.Join
  ( ColumnsMapping
  , JoinContext(..)
  , JoinMethod
  , Renamers
  , indexBy
  , leftJoin
  , rightJoin ) where

import Data.Map as Map (Map, insert, lookup)
import Data.Set as Set (Set, intersection, member, singleton)
import Data.Text as Text (Text)
import Data.Vector as Vector
  (Vector, cons, indexed, partitionWith, uncons, unsafeIndex, zipWith )

type Row = Vector Text
type Rows = Vector Row
type Document = (Row, Rows)
type ColumnsMapping = ([Text], [Text])
type Renamers = (Text -> Text, Text -> Text)

data JoinContext = JoinContext
  { columns :: ColumnsMapping
  , renamers :: Renamers }

type JoinMethod = JoinContext -> Rows -> Rows -> Rows

asSet :: (Foldable t, Ord a) => t a -> Set a
asSet = foldMap singleton

indexBy :: (Foldable t, Ord k) => (a -> k) -> t a -> Map k a
indexBy f = foldr (\a -> Map.insert (f a) a) mempty

select :: Vector Int -> Vector a -> Vector a
select mask values = unsafeIndex values <$> mask

indexMasks :: [Text] -> Vector Text -> (Vector Int, Vector Int)
indexMasks columns = partitionWith inKeys . indexed
  where
    keys = asSet columns
    inKeys (i, a) = (if a `member` keys then Left else Right) i

renameHeader :: (Foldable t, Functor t) => Renamers -> t Text -> t Text -> (t Text, t Text)
renameHeader (leftRenamer, rightRenamer) left right =
  (apply leftRenamer <$> left, apply rightRenamer <$> right)
  where
    common = intersection (asSet left) (asSet right)
    apply r c = if c `member` common then r c else c

filterRows :: Vector Int -> (Vector Int, Vector Int) -> Rows -> Rows -> Rows
filterRows mainKeyMask (sideKeyMask, sideValueMask) main side =
  (maybe emptyCells (select sideValueMask) . match) <$> mainKeys
  where
    mainKeys = select mainKeyMask <$> main
    optimize keys key | length main >= length side || key `member` keys = key
                      | otherwise = mempty
    indexedSide = indexBy (optimize (asSet mainKeys) . select sideKeyMask) side
    match = flip Map.lookup indexedSide
    emptyCells = (\_ -> "") <$> sideValueMask

joinInto :: JoinContext -> Maybe Document -> Document -> ((Row, Row), (Rows, Rows))
joinInto _ Nothing (header, rows) = ((header, mempty), (rows, mempty))
joinInto context (Just (sideHead, sideRows)) (mainHead, mainRows) =
  (newHead, (mainRows, filterRows mainKeyMask sideMasks mainRows sideRows))
  where
    (onLeft, onRight) = columns context
    mainKeyMask = fst (indexMasks onLeft mainHead)
    sideMasks@(_, valueMask) = indexMasks onRight sideHead
    newHead =
      renameHeader (renamers context) mainHead (select valueMask sideHead)

leftJoin :: JoinMethod
leftJoin context left right =
  build (joinInto context (Vector.uncons right) <$> Vector.uncons left)
  where
    build Nothing = mempty
    build (Just ((mainHead, sideHead), (mainRows, sideRows))) =
      Vector.cons (mainHead <> sideHead) (Vector.zipWith (<>) mainRows sideRows)

rightJoin :: JoinMethod
rightJoin context left right =
  build (joinInto context (Vector.uncons left) <$> Vector.uncons right)
  where
    build Nothing = mempty
    build (Just ((mainHead, sideHead), (mainRows, sideRows))) =
      Vector.cons (sideHead <> mainHead) (Vector.zipWith (<>) sideRows mainRows)
